from hyperpocket.tool import Tool
from hyperpocket.util.flatten_json_schema import flatten_json_schema


def tool_to_open_ai_spec(tool: Tool, use_profile: bool) -> dict:
    name = tool.name
    description = tool.get_description(use_profile=use_profile)
    arg_schema = tool.schema_model(use_profile=use_profile)
    json_schema = arg_schema.model_json_schema()
    json_schema = flatten_json_schema(json_schema)

    openai_spec = {
        "type": "function",
        "function": {
            "name": name,
            "description": description.strip(),
            "parameters": {**json_schema},
        },
    }

    return openai_spec



import ast
import json
import os
from unittest.async_case import IsolatedAsyncioTestCase

from openai import OpenAI
from pydantic import BaseModel

from hyperpocket.config import config
from hyperpocket_openai import PocketOpenAI


class TestPocketOpenAIUseProfile(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        config.public_server_port = "https"
        config.public_hostname = "localhost"
        config.public_server_port = 8001
        config.internal_server_port = 8000
        config.enable_local_callback_proxy = True

        self.pocket = PocketOpenAI(
            tools=[
                "https://github.com/vessl-ai/hyperpocket/main/tree/tools/slack/get-message",
                "https://github.com/vessl-ai/hyperpocket/main/tree/tools/slack/post-message",
                "https://github.com/vessl-ai/hyperpocket/main/tree/tools/none/simple-echo-tool",
                self.add,
                self.sub_pydantic_args,
            ],
            use_profile=True,
        )
        self.tool_specs = self.pocket.get_open_ai_tool_specs()
        self.client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))

    async def asyncTearDown(self):
        self.pocket._teardown_server()

    def test_get_tools_from_pocket_use_profile(self):
        # when
        specs = self.pocket.get_open_ai_tool_specs()
        get_tool, send_tool = specs[0], specs[1]

        # then
        self.assertIsInstance(get_tool, dict)
        self.assertEqual(get_tool["function"]["name"], "slack_get_messages")
        self.assertTrue(
            "channel"
            in get_tool["function"]["parameters"]["properties"]["body"]["properties"]
        )
        self.assertTrue(
            "limit"
            in get_tool["function"]["parameters"]["properties"]["body"]["properties"]
        )

        self.assertIsInstance(send_tool, dict)
        self.assertEqual(send_tool["function"]["name"], "slack_send_messages")
        self.assertTrue(
            "channel"
            in send_tool["function"]["parameters"]["properties"]["body"]["properties"]
        )
        self.assertTrue(
            "text"
            in send_tool["function"]["parameters"]["properties"]["body"]["properties"]
        )

    async def test_function_tool_use_profile(self):
        response = self.client.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "add 1, 2"}],
            tools=self.tool_specs,
        )

        choice = response.choices[0]
        tool_call = choice.message.tool_calls[0]
        name = tool_call.function.name
        args = tool_call.function.arguments
        args = json.loads(args)
        result = await self.pocket.ainvoke(tool_call)

        # then
        self.assertEqual(choice.finish_reason, "tool_calls")
        self.assertEqual(name, "add")
        self.assertEqual(args["body"], {"a": 1, "b": 2})
        self.assertEqual(result["content"], "3")

    async def test_pydantic_function_tool_use_profile(self):
        response = self.client.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "sub 1, 2"}],
            tools=self.tool_specs,
        )

        choice = response.choices[0]
        tool_call = choice.message.tool_calls[0]
        name = tool_call.function.name
        args = tool_call.function.arguments
        args = json.loads(args)
        result = await self.pocket.ainvoke(tool_call)

        # then
        self.assertEqual(choice.finish_reason, "tool_calls")
        self.assertEqual(name, "sub_pydantic_args")
        self.assertEqual(args["body"], {"a": {"first": 1}, "b": {"second": 2}})
        self.assertEqual(result["content"], "-1")

    async def test_wasm_tool_use_profile(self):
        response = self.client.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "echo 'hello world'"}],
            tools=self.tool_specs,
        )

        choice = response.choices[0]
        tool_call = choice.message.tool_calls[0]
        name = tool_call.function.name
        args = tool_call.function.arguments
        args = json.loads(args)
        result = await self.pocket.ainvoke(tool_call)
        output = ast.literal_eval(result["content"])

        # then
        self.assertEqual(choice.finish_reason, "tool_calls")
        self.assertEqual(name, "simple_echo_text")
        self.assertEqual(args["body"], {"text": "hello world"})
        self.assertTrue(output["stdout"].startswith("echo message : hello world"))

    @staticmethod
    def add(a: int, b: int) -> int:
        """
        Add two numbers

        Args:
            a(int): first number
            b(int): second number

        """

        return a + b

    class FirstNumber(BaseModel):
        first: int

    class SecondNumber(BaseModel):
        second: int

    @staticmethod
    def sub_pydantic_args(a: FirstNumber, b: SecondNumber):
        """
        sub two numbers

        Args:
            a(FirstNumber): first number
            b(SecondNumber): second number
        """
        return a.first - b.second



import asyncio
import json
from typing import List, Optional

from hyperpocket_openai.util import tool_to_open_ai_spec
from openai import OpenAI

try:
    from openai.types.chat import ChatCompletionMessageToolCall
except ImportError:
    raise ImportError("You need to install openai to use pocket PocketOpenAI.")

from hyperpocket import Pocket
from hyperpocket.config import pocket_logger
from hyperpocket.tool import Tool


class PocketOpenAI(Pocket):
    def invoke(self, tool_call: ChatCompletionMessageToolCall, **kwargs):
        loop = asyncio.get_running_loop()
        result = loop.run_until_complete(self.ainvoke(tool_call, **kwargs))
        return result

    async def ainvoke(self, tool_call: ChatCompletionMessageToolCall, **kwargs):
        arg_json = json.loads(tool_call.function.arguments)

        if self.use_profile:
            body = arg_json["body"]
            thread_id = arg_json.pop("thread_id", "default")
            profile = arg_json.pop("profile", "default")
        else:
            body = arg_json
            thread_id = "default"
            profile = "default"

        if isinstance(body, str):
            body = json.loads(body)

        result = await super().ainvoke(
            tool_call.function.name,
            body=body,
            thread_id=thread_id,
            profile=profile,
            **kwargs,
        )
        tool_message = {"role": "tool", "content": result, "tool_call_id": tool_call.id}

        return tool_message

    def get_open_ai_tool_specs(self, use_profile: Optional[bool] = None) -> List[dict]:
        if use_profile is not None:
            self.use_profile = use_profile
        specs = []
        for tool in self.core.tools.values():
            spec = self.get_open_ai_tool_spec(tool)
            specs.append(spec)
        return specs

    def get_open_ai_tool_spec(self, tool: Tool) -> dict:
        open_ai_spec = tool_to_open_ai_spec(tool, use_profile=self.use_profile)
        return open_ai_spec


async def handle_tool_call_async(
    llm: OpenAI,
    pocket: PocketOpenAI,
    model: str,
    tool_specs: List[dict],
    messages: List[dict],
):
    while True:
        response = llm.chat.completions.create(
            model=model,
            messages=messages,
            tools=tool_specs,
        )
        choice = response.choices[0]
        messages.append(choice.message)
        if choice.finish_reason == "stop":
            break

        elif choice.finish_reason == "tool_calls":
            tool_calls = choice.message.tool_calls
            for tool_call in tool_calls:
                pocket_logger.debug(f"[TOOL CALL] {tool_call}")
                tool_message = await pocket.ainvoke(tool_call)
                messages.append(tool_message)

    return messages[-1].content


def handle_tool_call(
    llm: OpenAI,
    pocket: PocketOpenAI,
    model: str,
    tool_specs: List[dict],
    messages: List[dict],
):
    while True:
        response = llm.chat.completions.create(
            model=model,
            messages=messages,
            tools=tool_specs,
        )
        choice = response.choices[0]
        messages.append(choice.message)
        if choice.finish_reason == "stop":
            break

        elif choice.finish_reason == "tool_calls":
            tool_calls = choice.message.tool_calls
            for tool_call in tool_calls:
                pocket_logger.debug("[TOOL CALL] ", tool_call)
                tool_message = pocket.invoke(tool_call)
                messages.append(tool_message)

    return messages[-1].content
